/*  This file is part of distributed Structured Prediction (dSP) - http://www.alexander-schwing.de/
*
*  distributed Structured Prediction (dSP) is free software: you can
*  redistribute it and/or modify it under the terms of the GNU General
*  Public License as published by the Free Software Foundation, either
*  version 3 of the License, or (at your option) any later version.
*
*  distributed Structured Prediction (dSP) is distributed in the hope
*  that it will be useful, but WITHOUT ANY WARRANTY; without even the
*  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
*  PURPOSE. See the GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with distributed Structured Prediction (dSP).
*  If not, see <http://www.gnu.org/licenses/>.
*
*  Copyright (C) 2010-2013  Alexander G. Schwing  [http://www.alexander-schwing.de/]
*/

//Author: Alexander G. Schwing

/*  ON A PERSONAL NOTE: I spent a significant amount of time to go through
*  both, theoretical justifications and coding of this framework.
*  I hope the package is useful for your task. Any requests,
*  feedback, donations and support would be greatly appreciated.
*  Thank you for contacting me!
*/

#include "libCRF.h"

#include <iostream>
#include <fstream>
#include <algorithm>
#include <map>
#include <cmath>
#include <limits>
#include <iostream>
#include <string.h>

#include <omp.h>

#include "../Test_libHCRF/GeneralDefinitions.h"

#include "../libBP/libBP.h"

#ifdef max
#undef max
#endif
#ifdef min
#undef min
#endif

//#define TIMING_OUTPUT
//#define THETA_PER_ITERATION

#define sign2(x) (( (x) > 0 ) - ( (x) < 0 ))

template <class T>
libCRF<T>::libCRF() {
	epsilon = T(1.0);
	p = T(2.0);
	C = T(1.0);
	rG = T(1e-5);
	max_iterations = 100;

	omp_set_num_threads(4);
}

template <class T>
libCRF<T>::~libCRF() {

}

template <class T>
int libCRF<T>::Initialize(std::vector<struct Feature2<T> > &FeatureArray, struct Feature2<T>* Loss, size_t totalTrainExamples) {

	this->FeatureArray = &FeatureArray;

	Vertices.resize(totalTrainExamples);
	Factors.resize(totalTrainExamples);
	Lambda.resize(totalTrainExamples);

	EdgeSet.resize(FeatureArray.size());

	for(size_t r=0;r<FeatureArray.size();++r) {
		EdgeSet[r].resize(totalTrainExamples);
		for(size_t x=0;x<FeatureArray[r].Vertices.size();++x) {
			for(typename std::map<size_t, struct VertexData<T> >::iterator v=FeatureArray[r].Vertices[x].begin();v!=FeatureArray[r].Vertices[x].end();++v) {
				Vertices[x][v->first].phi_r[r] = v->second.potData;
				Vertices[x][v->first].NumStates = v->second.NumStates;

				for(typename std::map<size_t, struct PotentialStruct<T>*>::iterator neighbors = v->second.ConnectedToFactor.begin();neighbors!=v->second.ConnectedToFactor.end();++neighbors) {
					Vertices[x][v->first].ConnectedTo.insert(std::pair<size_t, std::vector<T>*>(neighbors->first, NULL));

					//size_t tmp = neighbors->first;

					Factors[x][neighbors->first].phi_r.insert(std::pair<size_t, struct PotentialStruct<T>*>(r, neighbors->second));
					Factors[x][neighbors->first].lambda_v.insert(std::pair<size_t, std::vector<T>*>(v->first, NULL));

					EdgeSet[r][x].insert(std::pair<size_t, InternalFactorData<T>*>(neighbors->first, &Factors[x][neighbors->first]));
				}
			}
		}
	}

	if(Loss!=NULL) {
		for(size_t x=0;x<Loss->Vertices.size();++x) {
			for(typename std::map<size_t, struct VertexData<T> >::iterator v=Loss->Vertices[x].begin();v!=Loss->Vertices[x].end();++v) {
				Vertices[x][v->first].NodeLoss = v->second.potData;

				for(typename std::map<size_t, struct PotentialStruct<T>*>::iterator neighbors = v->second.ConnectedToFactor.begin();neighbors!=v->second.ConnectedToFactor.end();++neighbors) {
					Vertices[x][v->first].ConnectedTo.insert(std::pair<size_t, std::vector<T>*>(neighbors->first, NULL));

					Factors[x][neighbors->first].FactorLoss = neighbors->second;
				}
			}
		}
	}

	for(size_t x=0;x<totalTrainExamples;++x) {
		for(typename std::map<size_t,InternalNodeData<T> >::iterator v=Vertices[x].begin();v!=Vertices[x].end();++v) {
			for(typename std::map<size_t, std::vector<T>*>::iterator alpha = v->second.ConnectedTo.begin();alpha!=v->second.ConnectedTo.end();++alpha) {
				Lambda[x][v->first][alpha->first].assign(v->second.NumStates, T(0.0));
				alpha->second = &Lambda[x].find(v->first)->second.find(alpha->first)->second;
			}
		}
		for(typename std::map<size_t, InternalFactorData<T> >::iterator alpha=Factors[x].begin();alpha!=Factors[x].end();++alpha) {
			for(typename std::map<size_t, std::vector<T>*>::iterator v=alpha->second.lambda_v.begin();v!=alpha->second.lambda_v.end();++v) {
				v->second = &Lambda[x].find(v->first)->second.find(alpha->first)->second;
			}
		}
	}

	return 0;
}

template <class T>
int libCRF<T>::Clear() {
	std::vector<std::map<size_t, InternalNodeData<T> > >().swap(Vertices);
	std::vector<std::map<size_t, InternalFactorData<T> > >().swap(Factors);		//the second map, second element is a pointer to the message from this node to the factor
	std::vector<std::map<size_t, std::map<size_t, std::vector<T> > > >().swap(Lambda);
	std::vector<std::vector<std::map<size_t, InternalFactorData<T>* > > >().swap(EdgeSet);

	return 0;
}

template <class T>
int libCRF<T>::ResetMessages() {
	size_t totalTrainExamples = Vertices.size();
	for(size_t x=0;x<totalTrainExamples;++x) {
		for(typename std::map<size_t,InternalNodeData<T> >::iterator v=Vertices[x].begin();v!=Vertices[x].end();++v) {
			for(typename std::map<size_t, std::vector<T>*>::iterator alpha = v->second.ConnectedTo.begin();alpha!=v->second.ConnectedTo.end();++alpha) {
				alpha->second->assign(v->second.NumStates, T(0.0));
			}
		}
	}
	return 0;
}

template <class T>
T libCRF<T>::entropy(std::vector<T> &Input) {
	T retVal = T(0.0);
	typename std::vector<T>::const_iterator endPtr = Input.end();
	for(typename std::vector<T>::const_iterator iter=Input.begin();iter!=endPtr;++iter) {
		retVal += *iter*std::log(*iter+T(1e-100));
	}
	return -retVal;
}

template <class T>
T libCRF<T>::logNorm(std::vector<T> &Input) {
	T maxVal = Input[0];
	T res = T(0.0);
	typename std::vector<T>::const_iterator endPtr = Input.end();
	typename std::vector<T>::const_iterator beginPtr = Input.begin();
	for(typename std::vector<T>::const_iterator iter=beginPtr;iter!=endPtr;++iter) {
		maxVal = (*iter>maxVal) ? *iter : maxVal;
	}
	if(epsilon>0) {
		for(typename std::vector<T>::const_iterator iter=beginPtr;iter!=endPtr;++iter) {
			res += std::exp( (*iter - maxVal)/epsilon );
		}
		return maxVal + epsilon*std::log(res+T(1e-100));
	} else {
		return maxVal;
	}
}

template <class T>
int libCRF<T>::logNorm(std::vector<T> &FinalResult, const std::vector<T> &Input, size_t DimOfInterest, size_t numVars, const size_t* const num_states, const size_t* const stateProducts) {
	const size_t num_statesDOI = num_states[DimOfInterest];
	const size_t stateProductsNV = stateProducts[numVars];
	const size_t stateProductsDOI = stateProducts[DimOfInterest];
	
	FinalResult.assign(num_statesDOI, -std::numeric_limits<T>::max());

	std::vector<T> TmpVector(num_statesDOI, T(0.0));

	T *curVal;
	for(size_t k=0;k<stateProductsNV;++k) {
		curVal = &FinalResult[0] + (k/stateProductsDOI)%num_statesDOI;
		*curVal = (*curVal < Input[k]) ? Input[k] : *curVal;
	}

	if(epsilon > 0) {
		for(size_t k=0;k<stateProductsNV;++k) {
			const size_t pos = (k/stateProductsDOI)%num_statesDOI;
			TmpVector[pos] += std::exp( (Input[k] - FinalResult[pos])/epsilon );
		}
		for(size_t k=0;k<num_statesDOI;++k) {
			FinalResult[k] += epsilon*std::log(TmpVector[k]+T(1e-100));
		}
	}
	return 0;
}

template <class T>
T libCRF<T>::Iterate(std::vector<T> *theta, std::vector<T> &empiricalMeans) {
	size_t totalTrainExamples = Vertices.size();

	T oldPrimalValue = std::numeric_limits<T>::max();

#ifdef TIMING_OUTPUT
	CPrecisionTimer CTmr, CTmr2;
#endif
#ifdef THETA_PER_ITERATION
	CPrecisionTimer CThetaTmr, CThetaTmr2;
	std::ofstream thetaOFS("ThetaIteration.txt", std::ios_base::out | std::ios_base::binary);
	thetaOFS.write((char*)&max_iterations, sizeof(size_t));
	size_t thetaSize = theta->size();
	thetaOFS.write((char*)&thetaSize, sizeof(size_t));
	thetaOFS.write((char*)&(*theta)[0], sizeof(T)*theta->size());
#endif

	for(size_t iteration=0;iteration<max_iterations;++iteration) {

		/*if(iteration%1000==999) {
			char fn[101];
			sprintf(fn,"Theta_%d.txt",iteration);
			std::ofstream ofsTheta(fn, std::ios_base::out | std::ios_base::binary);
			ofsTheta.write((char*)&(*theta)[0], sizeof(T)*theta->size());
			ofsTheta.close();
		}*/

#ifdef	TIMING_OUTPUT
		CTmr.Start();
#endif
#ifdef THETA_PER_ITERATION
		CThetaTmr.Start();
		CThetaTmr2.Start();
#endif

		//message passing stuff
#pragma omp parallel
{
		std::vector<T> muArrays;
#pragma omp for nowait
		for(int x=0;x<(int)totalTrainExamples;++x) {
			typename std::map<size_t, InternalNodeData<T> >::const_iterator endPtr_v = Vertices[x].end();
			for(typename std::map<size_t, InternalNodeData<T> >::const_iterator v=Vertices[x].begin();v!=endPtr_v;++v) {

				size_t num_states = v->second.NumStates;//v->second.ConnectedTo.begin()->second->size();
				size_t betaIX = 0;
				size_t numNeighbors = v->second.ConnectedTo.size();

				std::vector<std::vector<T> > muArraysFinal(numNeighbors, std::vector<T>(num_states, T(0.0)));
				typename std::map<size_t, std::vector<T>*>::const_iterator endPtr_beta = v->second.ConnectedTo.end();
				for(typename std::map<size_t, std::vector<T>*>::const_iterator beta=v->second.ConnectedTo.begin();beta!=endPtr_beta;++beta, ++betaIX) {

					InternalFactorData<T>* FactorPtr = &Factors[x][beta->first];		// make a direct link
					struct PotentialStruct<T>* potBeginPtr = (FactorPtr->phi_r.size()>0) ? FactorPtr->phi_r.begin()->second : FactorPtr->FactorLoss;

					size_t numVars = potBeginPtr->num_variables;
					size_t* stateProducts = new size_t[numVars+1];
					size_t* num_states = new size_t[numVars];
					stateProducts[0] = 1;
					for(size_t k=0;k<numVars;++k) {
						num_states[k] = potBeginPtr->num_states[k];
						stateProducts[k+1] = stateProducts[k]*num_states[k];
					}

					//std::vector<T> muArrays(stateProducts[numVars],T(0.0));
					if(muArrays.size()==stateProducts[numVars]) {
						memset(&muArrays[0], 0, sizeof(T)*stateProducts[numVars]);
					} else {
						muArrays.assign(stateProducts[numVars], T(0.0));
					}
					/*if(muArrays.size()!=stateProducts[numVars]) {
						muArrays.assign(stateProducts[numVars], T(0.0));
					}*/

					if(FactorPtr->FactorLoss!=NULL) {
						T* endPtr_buf = FactorPtr->FactorLoss->pot + stateProducts[numVars];
						T* ptr_mu = &muArrays[0];
						for(T* buf = FactorPtr->FactorLoss->pot;buf!=endPtr_buf;++buf, ++ptr_mu) {
							*ptr_mu = *buf;
						}
					}

					typename std::map<size_t, struct PotentialStruct<T>*>::const_iterator endPtr_r = FactorPtr->phi_r.end();
					for(typename std::map<size_t, struct PotentialStruct<T>*>::const_iterator r=FactorPtr->phi_r.begin();r!=endPtr_r;++r) {
						T* endPtr_buf = r->second->pot + stateProducts[numVars];
						T* ptr_mu = &muArrays[0];
						//for(size_t k=0;k<stateProducts[numVars];++k) {
						for(T* buf = r->second->pot;buf!=endPtr_buf;++buf, ++ptr_mu) {
							//muArrays[k] += (*theta)[r->first]*r->second->pot[k];
							*ptr_mu += (*theta)[r->first]**buf;
						}
					}

					size_t hat_y_v_Index = -1;
					size_t neighborIX = 0;
					typename std::map<size_t, std::vector<T>*>::const_iterator endPtr_u = FactorPtr->lambda_v.end();
					for(typename std::map<size_t, std::vector<T>*>::const_iterator u=FactorPtr->lambda_v.begin();u!=endPtr_u;++u, ++neighborIX) {
						if(u->first!=v->first) {
							for(size_t k=0;k<stateProducts[numVars];++k) {
								muArrays[k] += (*u->second)[(k/stateProducts[neighborIX])%num_states[neighborIX]];
							}
						} else {
							hat_y_v_Index = neighborIX;
						}
					}

					logNorm(muArraysFinal[betaIX], muArrays, hat_y_v_Index, numVars, num_states, stateProducts);

					delete [] stateProducts;
					delete [] num_states;
				}

				std::vector<T> LambdaSum(numNeighbors,T(0.0));
				for(size_t hat_y_v=0;hat_y_v<num_states;++hat_y_v) {
					//computation of lambda
					T sumResult = (v->second.NodeLoss!=NULL) ? v->second.NodeLoss->pot[hat_y_v] : T(0.0);

					for(betaIX=0;betaIX<numNeighbors;++betaIX) {
						sumResult += muArraysFinal[betaIX][hat_y_v];
					}

					typename std::map<size_t, PotentialStruct<T>*>::const_iterator endPtr_r = v->second.phi_r.end();
					for(typename std::map<size_t, PotentialStruct<T>*>::const_iterator r=v->second.phi_r.begin();r!=endPtr_r;++r) {
						if(r->second!=NULL)
							sumResult += (*theta)[r->first]*r->second->pot[hat_y_v]; 
					}

					betaIX = 0;
					typename std::map<size_t, std::vector<T>*>::const_iterator endPtr_alpha = v->second.ConnectedTo.end();
					for(typename std::map<size_t, std::vector<T>*>::const_iterator alpha=v->second.ConnectedTo.begin();alpha!=endPtr_alpha;++alpha, ++betaIX) {
						//Lambda[x][v->first][alpha->first][hat_y_v] = sumResult/(1+v->second.ConnectedTo.size()) - muArraysFinal[betaIX][hat_y_v];
						//LambdaSum[betaIX] += Lambda[x][v->first][alpha->first][hat_y_v];
						(*alpha->second)[hat_y_v] = sumResult/(1+v->second.ConnectedTo.size()) - muArraysFinal[betaIX][hat_y_v];
						LambdaSum[betaIX] += (*alpha->second)[hat_y_v];
					}
				}

				for(betaIX=0;betaIX<numNeighbors;++betaIX) {
					LambdaSum[betaIX] /= num_states;
				}

				betaIX = 0;
				typename std::map<size_t, std::vector<T>*>::const_iterator endPtr_alpha = v->second.ConnectedTo.end();
				for(typename std::map<size_t, std::vector<T>*>::const_iterator alpha=v->second.ConnectedTo.begin();alpha!=endPtr_alpha;++alpha, ++betaIX) {
					for(size_t hat_y_v=0;hat_y_v<num_states;++hat_y_v) {
						//Lambda[x][v->first][alpha->first][hat_y_v] -= LambdaSum[betaIX];
						(*alpha->second)[hat_y_v] -= LambdaSum[betaIX];
					}
				}
			}
		}
}//end parallel section

#ifdef	TIMING_OUTPUT
		std::cout << "MSG: " << CTmr.Stop() << std::endl;
		CTmr.Start();
#endif

		//std::vector<T> locBel_x_v;
		//std::vector<T> facBel_x_alpha;

		//update theta
		std::vector<T> newTheta(*theta);
		for(size_t r=0;r<FeatureArray->size();++r) {
#ifdef	TIMING_OUTPUT
			std::cout << r;
			CTmr2.Start();
#endif

			T gradient = T(0.0);
#pragma omp parallel
{
			T gradient_threadLocal = T(0.0);
			std::vector<T> locBel_x_v;
			std::vector<T> facBel_x_alpha;
#pragma omp for nowait
			for(int x=0;x<int(totalTrainExamples);++x) {
				typename std::map<size_t, struct VertexData<T> >::const_iterator endPtr_v = (*FeatureArray)[r].Vertices[x].end();
				for(typename std::map<size_t, struct VertexData<T> >::const_iterator v=(*FeatureArray)[r].Vertices[x].begin();v!=endPtr_v;++v) {
					if(v->second.potData!=NULL) {
						computeLocalBelief(x, v->first, &locBel_x_v, *theta);
						//size_t numVars = locBel_x_v.size();
						T* potTmp = v->second.potData->pot;

						typename std::vector<T>::const_iterator endPtr_hat_y_v = locBel_x_v.end();
						for(typename std::vector<T>::const_iterator hat_y_v=locBel_x_v.begin();hat_y_v!=endPtr_hat_y_v;++hat_y_v, ++potTmp) {
							gradient_threadLocal += *hat_y_v**potTmp;
						}
					}
				}

				typename std::map<size_t, InternalFactorData<T>* >::const_iterator endPtr_alpha = EdgeSet[r][x].end();
				for(typename std::map<size_t, InternalFactorData<T>* >::const_iterator alpha = EdgeSet[r][x].begin();alpha!=endPtr_alpha;++alpha) {
					computeFactorBelief(x, alpha->second, &facBel_x_alpha, *theta);
					//size_t factorSize = facBel_x_alpha.size();
					T* potTmp = alpha->second->phi_r[r]->pot;

					typename std::vector<T>::const_iterator endPtr_hat_y_alpha = facBel_x_alpha.end();
					for(typename std::vector<T>::const_iterator hat_y_alpha=facBel_x_alpha.begin();hat_y_alpha!=endPtr_hat_y_alpha;++hat_y_alpha, ++potTmp) {
						gradient_threadLocal += *hat_y_alpha**potTmp;
					}
				}
			}
#pragma omp critical
{
			gradient += gradient_threadLocal;
}
}

#ifdef	TIMING_OUTPUT
			std::cout << "; Gradient: " << CTmr2.Stop();
			CTmr2.Start();
#endif

			gradient = gradient - empiricalMeans[r] + C*std::pow(std::abs((*theta)[r]), p-1)*sign2((*theta)[r]);
			T primalValue = computePrimalOfFeature(r, *theta, empiricalMeans);

#ifdef TIMING_OUTPUT
			std::cout << "; Value: " << gradient << "; Primal: " << primalValue;
#endif

			T stepsize = T(1.0);
			for(size_t armijoIteration=0;armijoIteration<100;++armijoIteration) {
				newTheta[r] = (*theta)[r] - stepsize*gradient;
				T newPrimalValue = computePrimalOfFeature(r, newTheta, empiricalMeans);
				if(newPrimalValue<primalValue) {
					(*theta)[r] = newTheta[r];
					break;
				}
				stepsize /= 2;
			}
			newTheta[r] = (*theta)[r];

#ifdef	TIMING_OUTPUT
			std::cout << "; Armijo: " << CTmr2.Stop() << " Step: " << stepsize << " New Theta: " << (*theta)[r] << std::endl;
#endif
		}
#ifdef	TIMING_OUTPUT
		std::cout << "Theta: " << CTmr.Stop() << std::endl;
		CTmr.Start();
#endif

#ifdef THETA_PER_ITERATION
		double time2 = CThetaTmr2.Stop();
#endif

		T primalValue = computePrimal(*theta, empiricalMeans);
#ifdef TIMING_OUTPUT
		std::cout << "Primal: " << primalValue << std::endl;
#endif
		
		bool outputBeliefs = false;
		T dualValue = computeDual(*theta, empiricalMeans, outputBeliefs);
		std::cout << iteration << " - Primal: " << primalValue << "\tDual: " << dualValue  << "\tGap:" << primalValue-dualValue << std::endl;

		if(oldPrimalValue < primalValue)
			std::cout << "primal Value error: " << primalValue - oldPrimalValue << "; iteration: " << iteration << std::endl;

#ifdef THETA_PER_ITERATION
		double time = CThetaTmr.Stop();
		thetaOFS.write((char*)&iteration, sizeof(size_t));
		thetaOFS.write((char*)&time, sizeof(double));
		thetaOFS.write((char*)&time2, sizeof(double));
		thetaOFS.write((char*)&primalValue, sizeof(T));
		thetaOFS.write((char*)&dualValue, sizeof(T));
		thetaOFS.write((char*)&(*theta)[0], sizeof(T)*theta->size());
#endif

		oldPrimalValue = primalValue;

		T relGap = (oldPrimalValue - dualValue)/oldPrimalValue;
		relGap = (relGap<0) ? -relGap : relGap;
		if(relGap < rG) {
			break;
		}

#ifdef	TIMING_OUTPUT
		std::cout << "Final: " << CTmr.Stop() << std::endl;
#endif
	}

#ifdef THETA_PER_ITERATION
	thetaOFS.close();
#endif

	T dualValue = computeDual(*theta, empiricalMeans, false);
	std::cout << "Primal: " << oldPrimalValue << "\tDual: " << dualValue << std::endl;

	return oldPrimalValue;
}

template<class T>
int libCRF<T>::computeLocalBelief(size_t x, size_t v_hat, std::vector<T> *locBel_x_v, std::vector<T> &theta) {
	InternalNodeData<T>* v = &Vertices[x][v_hat];
	size_t num_states = v->NumStates;

	locBel_x_v->resize(num_states);

	T maxVal = 0;
	for(size_t hat_y_v=0;hat_y_v<num_states;++hat_y_v) {
		T sum = (v->NodeLoss!=NULL) ? v->NodeLoss->pot[hat_y_v] : T(0.0);
		//T sum = T(0.0);

		typename std::map<size_t, struct PotentialStruct<T>*>::const_iterator endPtr_r = v->phi_r.end();
		for(typename std::map<size_t, struct PotentialStruct<T>*>::const_iterator r = v->phi_r.begin();r!=endPtr_r;++r) {
			if(r->second!=NULL)
				sum += theta[r->first]*r->second->pot[hat_y_v]; 
		}

		typename std::map<size_t, std::vector<T>*>::const_iterator endPtr_alpha = v->ConnectedTo.end();
		for(typename std::map<size_t, std::vector<T>*>::const_iterator alpha=v->ConnectedTo.begin();alpha!=endPtr_alpha;++alpha) {
			sum -= (*alpha->second)[hat_y_v];
		}

		(*locBel_x_v)[hat_y_v] = sum;
		maxVal = (hat_y_v==0) ? sum : ( (sum>maxVal) ? sum : maxVal );
	}

	T sumLocalBelief = T(0.0);
	typename std::vector<T>::iterator endPtr_hat_y_v = locBel_x_v->end();
	for(typename std::vector<T>::iterator hat_y_v=locBel_x_v->begin();hat_y_v!=endPtr_hat_y_v;++hat_y_v) {
		if(epsilon>0) {
			*hat_y_v = exp( (*hat_y_v - maxVal)/epsilon );
		} else {
			*hat_y_v = (*hat_y_v==maxVal) ? T(1.0) : T(0.0);
		}
		sumLocalBelief += *hat_y_v;
	}

	for(typename std::vector<T>::iterator hat_y_v=locBel_x_v->begin();hat_y_v!=endPtr_hat_y_v;++hat_y_v) {
		*hat_y_v /= sumLocalBelief;
	}
	return 0;
}

template <class T>
int libCRF<T>::computeFactorBelief(size_t x, InternalFactorData<T>* alpha, std::vector<T> *factorBelief_x_alpha, std::vector<T> &theta) {
	struct PotentialStruct<T>* potPtr = alpha->phi_r.begin()->second;

	size_t numVars = potPtr->num_variables;
	size_t* stateProducts = new size_t[numVars+1];
	size_t* num_states = new size_t[numVars];
	stateProducts[0] = 1;
	for(size_t k=0;k<numVars;++k) {
		num_states[k] = potPtr->num_states[k];
		stateProducts[k+1] = stateProducts[k]*num_states[k];
	}

	factorBelief_x_alpha->resize(stateProducts[numVars]);

	T maxVal = 0;
	for(size_t hat_y_alpha=0;hat_y_alpha<stateProducts[numVars];++hat_y_alpha) {
		T sum = (alpha->FactorLoss!=NULL) ? alpha->FactorLoss->pot[hat_y_alpha] : T(0.0);

		typename std::map<size_t, struct PotentialStruct<T>*>::const_iterator endPtr_r = alpha->phi_r.end();
		for(typename std::map<size_t, struct PotentialStruct<T>*>::const_iterator r = alpha->phi_r.begin();r!=endPtr_r;++r) {
			sum += theta[r->first]*r->second->pot[hat_y_alpha];
		}

		size_t variableIX = 0;
		typename std::map<size_t, std::vector<T>*>::const_iterator endPtr_v = alpha->lambda_v.end();
		for(typename std::map<size_t, std::vector<T>*>::const_iterator v=alpha->lambda_v.begin();v!=endPtr_v;++v, ++variableIX) {
			sum += (*v->second)[ (hat_y_alpha/stateProducts[variableIX])%num_states[variableIX] ];
		}

		(*factorBelief_x_alpha)[hat_y_alpha] = sum;
		maxVal = (hat_y_alpha==0) ? sum : ( (sum>maxVal) ? sum : maxVal );
	}

	T sumFactorBelief = T(0.0);
	typename std::vector<T>::iterator endPtr_hat_y_alpha = factorBelief_x_alpha->end();
	for(typename std::vector<T>::iterator hat_y_alpha=factorBelief_x_alpha->begin();hat_y_alpha!=endPtr_hat_y_alpha;++hat_y_alpha) {
		if(epsilon>0) {
			*hat_y_alpha = exp( (*hat_y_alpha - maxVal)/epsilon );
		} else {
			*hat_y_alpha = (*hat_y_alpha==maxVal) ? T(1.0) : T(0.0);
		}
		sumFactorBelief += *hat_y_alpha;
	}

	for(typename std::vector<T>::iterator hat_y_alpha=factorBelief_x_alpha->begin();hat_y_alpha!=endPtr_hat_y_alpha;++hat_y_alpha) {
		*hat_y_alpha /= sumFactorBelief;
	}

	delete [] stateProducts;
	delete [] num_states;				

	return 0;
}

template<class T>
int libCRF<T>::computeBeliefs(std::vector<std::map<size_t, std::vector<T> > > &localBelief, std::vector<std::map<size_t, std::vector<T> > > &factorBelief, std::vector<T> &theta) {
	size_t totalTrainExamples = Vertices.size();

	//compute beliefs
#pragma omp parallel
{
#pragma omp for nowait
	for(int x=0;x<int(totalTrainExamples);++x) {
		//local beliefs
		for(typename std::map<size_t, InternalNodeData<T> >::iterator v=Vertices[x].begin();v!=Vertices[x].end();++v) {
			
			computeLocalBelief(x, v->first, &localBelief[x][v->first], theta);		//not beneficial as we can safe log(n) access time of Vertices[x][v_hat] in the computeLocalBelief function by passing the iterator!!!

		}

		//factor beliefs
		for(typename std::map<size_t, InternalFactorData<T> >::iterator alpha=Factors[x].begin();alpha!=Factors[x].end();++alpha) {
			
			computeFactorBelief(x, &alpha->second, &factorBelief[x][alpha->first], theta);
							
		}
	}
}//end parallel section

	return 0;
}

template <class T>
T libCRF<T>::computeDual(std::vector<T> &theta, std::vector<T> &empiricalMeans, bool outputBeliefs) {
	size_t totalTrainExamples = Vertices.size();

	std::vector<std::map<size_t, std::vector<T> > > localBelief(totalTrainExamples);
	std::vector<std::map<size_t, std::vector<T> > > factorBelief(totalTrainExamples);

	computeBeliefs(localBelief, factorBelief, theta);

	if(outputBeliefs) {
		std::cout << "Local Beliefs:" << std::endl;
		for(size_t x=0;x<totalTrainExamples;++x) {
			typename std::map<size_t, std::vector<T> >::const_iterator endPtr_v = localBelief[x].end();
			for(typename std::map<size_t, std::vector<T> >::const_iterator v=localBelief[x].begin();v!=endPtr_v;++v) {
				for(typename std::vector<T>::const_iterator iter = v->second.begin();iter!=v->second.end();++iter) {
					std::cout << *iter << "\t";
				}
				std::cout << std::endl;
			}
			std::cout << std::endl << std::endl;
		}

		std::cout << "Factor Beliefs:" << std::endl;
		for(size_t x=0;x<totalTrainExamples;++x) {
			typename std::map<size_t, std::vector<T> >::const_iterator endPtr_alpha = factorBelief[x].end();
			for(typename std::map<size_t, std::vector<T> >::const_iterator alpha=factorBelief[x].begin();alpha!=endPtr_alpha;++alpha) {
				for(typename std::vector<T>::const_iterator iter = alpha->second.begin();iter!=alpha->second.end();++iter) {
					std::cout << *iter << "\t";
				}
				std::cout << std::endl;
			}
			std::cout << std::endl << std::endl;
		}
	}

	T dualValue = T(0.0);
	T q = p/(p-1);
	T cHat = std::pow(C, 1-q);

#pragma omp parallel
{
	T dualValue_threadLocal = T(0.0);
#pragma omp for nowait
	for(int x=0;x<int(totalTrainExamples);++x) {
		typename std::map<size_t, InternalNodeData<T> >::const_iterator endPtr_v = Vertices[x].end();
		for(typename std::map<size_t, InternalNodeData<T> >::const_iterator v=Vertices[x].begin();v!=endPtr_v;++v) {
			std::vector<T>& locBel_v = localBelief[x][v->first];

			dualValue_threadLocal += epsilon*entropy(locBel_v);
			if(v->second.NodeLoss!=NULL) {
				T* pot = v->second.NodeLoss->pot;
				for(size_t y_v=0;y_v<v->second.NumStates;++y_v) {
					dualValue_threadLocal += locBel_v[y_v]*pot[y_v];
				}
			}
		}

		typename std::map<size_t, InternalFactorData<T> >::const_iterator endPtr_alpha = Factors[x].end();
		for(typename std::map<size_t, InternalFactorData<T> >::const_iterator alpha=Factors[x].begin();alpha!=endPtr_alpha;++alpha) {
			std::vector<T>& facBel_alpha = factorBelief[x][alpha->first];
			
			dualValue_threadLocal += epsilon*entropy(facBel_alpha);
			if(alpha->second.FactorLoss!=NULL) {
				T* pot = alpha->second.FactorLoss->pot;
				size_t absyalpha = facBel_alpha.size();
				for(size_t y_alpha=0;y_alpha<absyalpha;++y_alpha) {
					dualValue_threadLocal += facBel_alpha[y_alpha]*pot[y_alpha];
				}
			}
		}
	}
#pragma omp critical
{
	dualValue += dualValue_threadLocal;
}//end critical section
}//end parallel section

#pragma omp parallel
{
	T dualValue_threadLocal = T(0.0);
#pragma omp for nowait
	for(int r=0;r<int(FeatureArray->size());++r) {
		T z_r = T(0.0);
		for(size_t x=0;x<totalTrainExamples;++x) {
			typename std::map<size_t, struct VertexData<T> >::const_iterator endPtr_v = (*FeatureArray)[r].Vertices[x].end();
			for(typename std::map<size_t, struct VertexData<T> >::const_iterator v=(*FeatureArray)[r].Vertices[x].begin();v!=endPtr_v;++v) {
				if(v->second.potData!=NULL) {
					
					size_t numVars = v->second.NumStates;
					T* locBel_x_v = &localBelief[x][v->first][0];
					T* potTmp = v->second.potData->pot;
					for(size_t hat_y_v=0;hat_y_v<numVars;++hat_y_v) {
						z_r += locBel_x_v[hat_y_v]*potTmp[hat_y_v];
					}
				}
			}

			typename std::map<size_t, InternalFactorData<T>* >::const_iterator endPtr_alpha = EdgeSet[r][x].end();
			for(typename std::map<size_t, InternalFactorData<T>* >::const_iterator alpha = EdgeSet[r][x].begin();alpha!=endPtr_alpha;++alpha) {
				std::vector<T>* tmpFactor = &factorBelief[x][alpha->first];
				size_t factorSize = tmpFactor->size();

				T* facBel_x_alpha = &(*tmpFactor)[0];
				T* potTmp = alpha->second->phi_r[r]->pot;
				for(size_t hat_y_alpha=0;hat_y_alpha<factorSize;++hat_y_alpha) {
					z_r += facBel_x_alpha[hat_y_alpha]*potTmp[hat_y_alpha];
				}
			}
		}

		if(outputBeliefs) {
			std::cout << z_r << "\t" << empiricalMeans[r] << std::endl;
		}

		dualValue_threadLocal -= cHat/q*std::pow(std::abs(z_r - empiricalMeans[r]), q);
	}
#pragma omp critical
{
	dualValue += dualValue_threadLocal;
}//end critical section
}//end parallel section

	return dualValue;
}

template <class T>
T libCRF<T>::computePrimalOfFeature(size_t cur_r, std::vector<T> &theta, std::vector<T> &empiricalMeans) {
	size_t totalTrainExamples = Vertices.size();

	T primalValue = T(0.0);
	//compute beliefs
#pragma omp parallel
{
	T threadLoc_primalValue = T(0.0);
	std::vector<T> logBelief;
#pragma omp for nowait
	for(int x=0;x<(int)totalTrainExamples;++x) {
		//local beliefs
		typename std::map<size_t, struct VertexData<T> >::const_iterator endPtr_v = (*FeatureArray)[cur_r].Vertices[x].end();
		for(typename std::map<size_t, struct VertexData<T> >::const_iterator v=(*FeatureArray)[cur_r].Vertices[x].begin();v!=endPtr_v;++v) {
			InternalNodeData<T>* v2 = &Vertices[x][v->first];
			size_t num_states = v->second.NumStates;
			logBelief.resize(num_states);

			for(size_t hat_y_v=0;hat_y_v<num_states;++hat_y_v) {
				T sum = (v2->NodeLoss!=NULL) ? v2->NodeLoss->pot[hat_y_v] : T(0.0);
				//T sum = T(0.0);

				typename std::map<size_t, struct PotentialStruct<T>*>::const_iterator endPtr_r = v2->phi_r.end();
				for(typename std::map<size_t, struct PotentialStruct<T>*>::const_iterator r = v2->phi_r.begin();r!=endPtr_r;++r) {
					if(r->second!=NULL)
						sum += theta[r->first]*r->second->pot[hat_y_v]; 
				}

				typename std::map<size_t, std::vector<T>*>::const_iterator endPtr_alpha = v2->ConnectedTo.end();
				for(typename std::map<size_t, std::vector<T>*>::const_iterator alpha=v2->ConnectedTo.begin();alpha!=endPtr_alpha;++alpha) {
					sum -= (*alpha->second)[hat_y_v];
				}

				logBelief[hat_y_v] = sum;
			}

			threadLoc_primalValue += logNorm(logBelief);
		}

		//factor beliefs
		typename std::map<size_t, InternalFactorData<T>* >::const_iterator endPtr_alpha = EdgeSet[cur_r][x].end();
		for(typename std::map<size_t, InternalFactorData<T>* >::const_iterator alpha = EdgeSet[cur_r][x].begin();alpha!=endPtr_alpha;++alpha) {
			struct PotentialStruct<T>* potPtr = alpha->second->phi_r.begin()->second;
			size_t numVars = potPtr->num_variables;
			size_t* stateProducts = new size_t[numVars+1];
			size_t* num_states = new size_t[numVars];
			stateProducts[0] = 1;
			size_t* tmp_ptr = potPtr->num_states;
			for(size_t k=0;k<numVars;++k) {
				num_states[k] = tmp_ptr[k];
				stateProducts[k+1] = stateProducts[k]*num_states[k];
			}

			logBelief.resize(stateProducts[numVars]);

			for(size_t hat_y_alpha=0;hat_y_alpha<stateProducts[numVars];++hat_y_alpha) {
				T sum = (alpha->second->FactorLoss!=NULL) ? alpha->second->FactorLoss->pot[hat_y_alpha] : T(0.0);
				//T sum = T(0.0);

				typename std::map<size_t, struct PotentialStruct<T>*>::const_iterator endPtr_r = alpha->second->phi_r.end();
				for(typename std::map<size_t, struct PotentialStruct<T>*>::const_iterator r = alpha->second->phi_r.begin();r!=endPtr_r;++r) {
					sum += theta[r->first]*r->second->pot[hat_y_alpha];
				}

				size_t variableIX = 0;
				typename std::map<size_t, std::vector<T>*>::const_iterator endPtr_v = alpha->second->lambda_v.end();
				for(typename std::map<size_t, std::vector<T>*>::const_iterator v=alpha->second->lambda_v.begin();v!=endPtr_v;++v, ++variableIX) {
					sum += (*v->second)[ (hat_y_alpha/stateProducts[variableIX])%num_states[variableIX] ];
				}

				logBelief[hat_y_alpha] = sum;
			}

			threadLoc_primalValue += logNorm(logBelief);

			delete [] stateProducts;
			delete [] num_states;
		}
	}
#pragma omp critical
{
	primalValue += threadLoc_primalValue;
}
}

	primalValue -= theta[cur_r]*empiricalMeans[cur_r];
	primalValue += (C/p)*std::pow(std::abs(theta[cur_r]),p);


	return primalValue;
}

template <class T>
T libCRF<T>::computePrimal(std::vector<T> &theta, std::vector<T> &empiricalMeans) {
	size_t totalTrainExamples = Vertices.size();

	T primalValue = T(0.0);
	//compute beliefs
#pragma omp parallel
{
	T threadLoc_primalValue = T(0.0);
	std::vector<T> logBelief;
#pragma omp for nowait
	for(int x=0;x<(int)totalTrainExamples;++x) {
		//local beliefs
		typename std::map<size_t, InternalNodeData<T> >::const_iterator endPtr_v = Vertices[x].end();
		for(typename std::map<size_t, InternalNodeData<T> >::const_iterator v=Vertices[x].begin();v!=endPtr_v;++v) {
			size_t num_states = v->second.NumStates;
			logBelief.resize(num_states);

			for(size_t hat_y_v=0;hat_y_v<num_states;++hat_y_v) {
				T sum = (v->second.NodeLoss!=NULL) ? v->second.NodeLoss->pot[hat_y_v] : T(0.0);
				//T sum = T(0.0);

				typename std::map<size_t, struct PotentialStruct<T>*>::const_iterator endPtr_r = v->second.phi_r.end();
				for(typename std::map<size_t, struct PotentialStruct<T>*>::const_iterator r = v->second.phi_r.begin();r!=endPtr_r;++r) {
					if(r->second!=NULL)
						sum += theta[r->first]*r->second->pot[hat_y_v]; 
				}

				typename std::map<size_t, std::vector<T>*>::const_iterator endPtr_alpha = v->second.ConnectedTo.end();
				for(typename std::map<size_t, std::vector<T>*>::const_iterator alpha=v->second.ConnectedTo.begin();alpha!=endPtr_alpha;++alpha) {
					sum -= (*alpha->second)[hat_y_v];
				}

				logBelief[hat_y_v] = sum;
			}

			threadLoc_primalValue += logNorm(logBelief);
		}

		//factor beliefs
		typename std::map<size_t, InternalFactorData<T> >::const_iterator endPtr_alpha = Factors[x].end();
		for(typename std::map<size_t, InternalFactorData<T> >::const_iterator alpha=Factors[x].begin();alpha!=endPtr_alpha;++alpha) {
			struct PotentialStruct<T>* potPtr = alpha->second.phi_r.begin()->second;
			size_t numVars = potPtr->num_variables;
			size_t* stateProducts = new size_t[numVars+1];
			size_t* num_states = new size_t[numVars];
			stateProducts[0] = 1;
			size_t* tmp_ptr = potPtr->num_states;
			for(size_t k=0;k<numVars;++k) {
				num_states[k] = tmp_ptr[k];
				stateProducts[k+1] = stateProducts[k]*num_states[k];
			}

			logBelief.resize(stateProducts[numVars]);

			for(size_t hat_y_alpha=0;hat_y_alpha<stateProducts[numVars];++hat_y_alpha) {
				T sum = (alpha->second.FactorLoss!=NULL) ? alpha->second.FactorLoss->pot[hat_y_alpha] : T(0.0);
				//T sum = T(0.0);

				typename std::map<size_t, struct PotentialStruct<T>*>::const_iterator endPtr_r = alpha->second.phi_r.end();
				for(typename std::map<size_t, struct PotentialStruct<T>*>::const_iterator r = alpha->second.phi_r.begin();r!=endPtr_r;++r) {
					sum += theta[r->first]*r->second->pot[hat_y_alpha];
				}

				size_t variableIX = 0;
				typename std::map<size_t, std::vector<T>*>::const_iterator endPtr_v = alpha->second.lambda_v.end();
				for(typename std::map<size_t, std::vector<T>*>::const_iterator v=alpha->second.lambda_v.begin();v!=endPtr_v;++v, ++variableIX) {
					sum += (*v->second)[ (hat_y_alpha/stateProducts[variableIX])%num_states[variableIX] ];
				}

				logBelief[hat_y_alpha] = sum;
			}

			threadLoc_primalValue += logNorm(logBelief);

			delete [] stateProducts;
			delete [] num_states;				
		}
	}
#pragma omp critical
{
	primalValue += threadLoc_primalValue;
}
}

	for(size_t r=0;r<theta.size();++r) {
		primalValue -= theta[r]*empiricalMeans[r];
		primalValue += (C/p)*std::pow(std::abs(theta[r]),p);
	}

	return primalValue;
}

template <class T>
double libCRF<T>::ComputeMAP(std::vector<T> *theta, std::vector<std::map<size_t, std::vector<T> > >& AllLocalBeliefs, const char* folder = NULL) {
	size_t totalSamples = Vertices.size();

	AllLocalBeliefs.resize(totalSamples);
	double inferenceTime = 0;

#pragma omp parallel
{
	double inferenceTime_threadLocal = 0;
	CPrecisionTimer CTmr;
#pragma omp for nowait schedule(dynamic,1)
	for(int x=0;x<(int)totalSamples;++x) {

		std::map<size_t, std::pair<struct PotentialStruct<T>, std::set< size_t > > > LocalBPPot;
		std::map<size_t, std::pair<struct PotentialStruct<T>, std::set< size_t > > > FactorBPPot;

		//for all vertices
		for(typename std::map<size_t, InternalNodeData<T> >::iterator v=Vertices[x].begin();v!=Vertices[x].end();++v) {
			//size_t numStates = v->second.phi_r.begin()->second->num_states[0];
			size_t numStates = v->second.NumStates;
			for(typename std::map<size_t, struct PotentialStruct<T>*>::iterator r=v->second.phi_r.begin();r!=v->second.phi_r.end();++r) {
				if(r->second!=NULL) {
					if(LocalBPPot[v->first].first.pot==NULL) {
						T *tmp = new T[numStates];
						memset((void*)tmp, 0, numStates*sizeof(T));
						LocalBPPot[v->first].first.pot = tmp;
						LocalBPPot[v->first].first.num_states = new size_t[1];
						LocalBPPot[v->first].first.num_states[0] = numStates;
						LocalBPPot[v->first].first.num_variables = 1;
					}
					for(size_t h=0;h<numStates;++h) {
						LocalBPPot[v->first].first.pot[h] += theta->at(r->first)*r->second->pot[h];
					}
				}
			}
		}

		//for all factors
		for(typename std::map<size_t, InternalFactorData<T> >::iterator alpha=Factors[x].begin();alpha!=Factors[x].end();++alpha) {
			size_t numVars = alpha->second.phi_r.begin()->second->num_variables;
			size_t* stateProducts = new size_t[numVars+1];

			stateProducts[0] = 1;
			for(size_t k=0;k<numVars;++k) {
				stateProducts[k+1] = stateProducts[k]*alpha->second.phi_r.begin()->second->num_states[k];
			}

			for(typename std::map<size_t, struct PotentialStruct<T>*>::iterator r=alpha->second.phi_r.begin();r!=alpha->second.phi_r.end();++r) {
				if(FactorBPPot[alpha->first].first.pot==NULL) {
					T *tmp = new T[stateProducts[numVars]];
					memset((void*)tmp, 0, stateProducts[numVars]*sizeof(T));
					FactorBPPot[alpha->first].first.pot = tmp;
					FactorBPPot[alpha->first].first.num_states = r->second->num_states;
					FactorBPPot[alpha->first].first.num_variables = numVars;
				}
				for(size_t y_alpha_H_alpha=0;y_alpha_H_alpha<stateProducts[numVars];++y_alpha_H_alpha) {
					FactorBPPot[alpha->first].first.pot[y_alpha_H_alpha] += theta->at(r->first)*r->second->pot[y_alpha_H_alpha];
				}
			}

			for(typename std::map<size_t, std::vector<T>*>::iterator u=alpha->second.lambda_v.begin();u!=alpha->second.lambda_v.end();++u) {
				FactorBPPot[alpha->first].second.insert(u->first);
				LocalBPPot[u->first].second.insert(alpha->first);

				if(LocalBPPot[u->first].first.pot==NULL) {
					LocalBPPot[u->first].first.num_variables = 1;
					LocalBPPot[u->first].first.num_states = new size_t[1];
					LocalBPPot[u->first].first.num_states[0] = u->second->size();
				}
			}

			delete [] stateProducts;
		}

		if(folder==NULL) {
			//for debugging: epsilon = 1.0
			libBP<T> BPAlgo;
			BPAlgo.Initialize(LocalBPPot, FactorBPPot, T(0.0));
			CTmr.Start();
			BPAlgo.RunBP(10000, true);
			inferenceTime_threadLocal += CTmr.Stop();
			BPAlgo.GetResult(AllLocalBeliefs[x]);
		} else {
			char fn[1025];
			sprintf(fn, "%sout_%d.uai", folder, x);
			WriteUAIInferenceFile(fn, LocalBPPot, FactorBPPot);
		}

		//clean up
		for(typename std::map<size_t, std::pair<struct PotentialStruct<T>, std::set< size_t > > >::iterator iter=LocalBPPot.begin();iter!=LocalBPPot.end();++iter) {
			delete [] iter->second.first.pot;
			delete [] iter->second.first.num_states;
		}
		for(typename std::map<size_t, std::pair<struct PotentialStruct<T>, std::set< size_t > > >::iterator iter=FactorBPPot.begin();iter!=FactorBPPot.end();++iter) {
			delete [] iter->second.first.pot;
		}
		LocalBPPot.clear();
		FactorBPPot.clear();
		std::map<size_t, std::pair<struct PotentialStruct<T>, std::set< size_t > > >().swap(LocalBPPot);
		std::map<size_t, std::pair<struct PotentialStruct<T>, std::set< size_t > > >().swap(FactorBPPot);

	}
#pragma omp critical
{
	inferenceTime += inferenceTime_threadLocal;
}//end critical section
}//end parallel section


	return inferenceTime/totalSamples;
}

template<class T>
int libCRF<T>::WriteUAIInferenceFile(const char* fn, std::map<size_t,std::pair<struct PotentialStruct<T>,std::set<size_t> > > &LocalPot, std::map<size_t,std::pair<struct PotentialStruct<T>,std::set<size_t> > > &FactorPot) {
	std::map<size_t,size_t> InternalToExternalNodeIndex;
	
	std::ofstream UAIFile(fn, std::ios_base::out);

	UAIFile << "MARKOV" << std::endl;

	UAIFile << LocalPot.size() << std::endl;

	size_t vIX = 0;
	size_t localPotCount = 0;
	for(typename std::map<size_t,std::pair<struct PotentialStruct<T>,std::set<size_t> > >::iterator v=LocalPot.begin();v!=LocalPot.end();++v) {
		UAIFile << v->second.first.num_states[0] << " ";
		InternalToExternalNodeIndex[v->first] = vIX++;
		localPotCount += (v->second.first.pot!=NULL) ? 1 : 0;
	}
	UAIFile << std::endl;

	UAIFile << FactorPot.size() + localPotCount << std::endl;

	for(typename std::map<size_t,std::pair<struct PotentialStruct<T>,std::set<size_t> > >::iterator v=LocalPot.begin();v!=LocalPot.end();++v) {
		if(v->second.first.pot!=NULL) {
			UAIFile << v->second.first.num_variables << " " << InternalToExternalNodeIndex[v->first] << std::endl;
		}
	}

	for(typename std::map<size_t,std::pair<struct PotentialStruct<T>,std::set<size_t> > >::iterator alpha=FactorPot.begin();alpha!=FactorPot.end();++alpha) {
		UAIFile << alpha->second.first.num_variables;
		for(std::set<size_t>::iterator v=alpha->second.second.begin();v!=alpha->second.second.end();++v) {
			UAIFile << " " << InternalToExternalNodeIndex[*v];
		}
		UAIFile << std::endl;
	}

	UAIFile << std::endl;
	for(typename std::map<size_t,std::pair<struct PotentialStruct<T>,std::set<size_t> > >::iterator v=LocalPot.begin();v!=LocalPot.end();++v) {
		if(v->second.first.pot!=NULL) {
			size_t sz = v->second.first.num_states[0];
			UAIFile << sz << std::endl;

			for(size_t y_v=0;y_v<sz;++y_v) {
				UAIFile << v->second.first.pot[y_v] << " ";
			}

			UAIFile << std::endl;
		}
	}

	for(typename std::map<size_t,std::pair<struct PotentialStruct<T>,std::set<size_t> > >::iterator alpha=FactorPot.begin();alpha!=FactorPot.end();++alpha) {
		size_t sz = 1;
		size_t numVars = alpha->second.first.num_variables;
		for(size_t k=0;k<numVars;++k) {
			sz *= alpha->second.first.num_states[k];
		}
		UAIFile << sz << std::endl;

		size_t* prodState = new size_t[numVars+1];
		size_t* prodStateRev = new size_t[numVars+1];
		prodState[numVars] = 1;
		prodStateRev[0] = 1;
		for(size_t n=numVars-1;n>=0 && n<numVars;--n) {
			prodState[n] = prodState[n+1]*alpha->second.first.num_states[n];
			int revVar = -int(n)+int(numVars);
			prodStateRev[revVar] = prodStateRev[revVar-1]*alpha->second.first.num_states[revVar-1];
		}

		size_t* varState = new size_t[numVars];

		for(size_t y_alpha=0;y_alpha<sz;++y_alpha) {
			for(size_t n=numVars-1;n>=0 && n<numVars;--n) {
				varState[n] = (y_alpha/prodState[n+1])%alpha->second.first.num_states[n];
			}

			size_t newPos = 0;
			for(size_t n=0;n<numVars;++n) {
				newPos += varState[n]*prodStateRev[n];
			}

			UAIFile << alpha->second.first.pot[newPos] << " ";
		}

		delete [] varState;
		delete [] prodState;
		delete [] prodStateRev;

		UAIFile << std::endl;
	}
	

	UAIFile.close();
	return 0;
}

template class libCRF<double>;
template class libCRF<float>;